@@ -4,7 +4,7 @@ use futures_lite::future::block_on;
 use futures_lite::{future::Boxed, FutureExt};
 use hyper::client::HttpConnector;
 use hyper::header::HeaderValue;
-use hyper::{body, body::Buf, client, header, Body, Client, Method, Request, Response, Uri};
+use hyper::{body, body::Buf, header, Body, Client, Method, Request, Response, Uri};
 use hyper_rustls::HttpsConnector;
 use libflate::gzip;
 use log::{error, trace, warn};
@@ -19,20 +19,21 @@ use std::{io, result::Result};
 use crate::dbg_msg;
 use crate::oauth::{force_refresh_token, token_daemon, Oauth};
 use crate::server::RequestExt;
-use crate::utils::format_url;
+use crate::utils::{format_url, Post};
 
 const REDDIT_URL_BASE: &str = "https://oauth.reddit.com";
-const ALTERNATIVE_REDDIT_URL_BASE: &str = "https://www.reddit.com";
+const REDDIT_URL_BASE_HOST: &str = "oauth.reddit.com";
 
-pub static CLIENT: Lazy<Client<HttpsConnector<HttpConnector>>> = Lazy::new(|| {
-	let https = hyper_rustls::HttpsConnectorBuilder::new()
-		.with_native_roots()
-		.expect("No native root certificates found")
-		.https_only()
-		.enable_http1()
-		.build();
-	client::Client::builder().build(https)
-});
+const REDDIT_SHORT_URL_BASE: &str = "https://redd.it";
+const REDDIT_SHORT_URL_BASE_HOST: &str = "redd.it";
+
+const ALTERNATIVE_REDDIT_URL_BASE: &str = "https://www.reddit.com";
+const ALTERNATIVE_REDDIT_URL_BASE_HOST: &str = "www.reddit.com";
+
+pub static HTTPS_CONNECTOR: Lazy<HttpsConnector<HttpConnector>> =
+	Lazy::new(|| hyper_rustls::HttpsConnectorBuilder::new().with_native_roots().https_only().enable_http2().build());
+
+pub static CLIENT: Lazy<Client<HttpsConnector<HttpConnector>>> = Lazy::new(|| Client::builder().build::<_, Body>(HTTPS_CONNECTOR.clone()));
 
 pub static OAUTH_CLIENT: Lazy<ArcSwap<Oauth>> = Lazy::new(|| {
 	let client = block_on(Oauth::new());
@@ -44,6 +45,11 @@ pub static OAUTH_RATELIMIT_REMAINING: AtomicU16 = AtomicU16::new(99);
 
 pub static OAUTH_IS_ROLLING_OVER: AtomicBool = AtomicBool::new(false);
 
+const URL_PAIRS: [(&str, &str); 2] = [
+	(ALTERNATIVE_REDDIT_URL_BASE, ALTERNATIVE_REDDIT_URL_BASE_HOST),
+	(REDDIT_SHORT_URL_BASE, REDDIT_SHORT_URL_BASE_HOST),
+];
+
 /// Gets the canonical path for a resource on Reddit. This is accomplished by
 /// making a `HEAD` request to Reddit at the path given in `path`.
 ///
@@ -57,13 +63,32 @@ pub static OAUTH_IS_ROLLING_OVER: AtomicBool = AtomicBool::new(false);
 /// `Location` header. An `Err(String)` is returned if Reddit responds with a
 /// 429, or if we were unable to decode the value in the `Location` header.
 #[cached(size = 1024, time = 600, result = true)]
-pub async fn canonical_path(path: String) -> Result<Option<String>, String> {
-	let res = reddit_head(path.clone(), true).await?;
+#[async_recursion::async_recursion]
+pub async fn canonical_path(path: String, tries: i8) -> Result<Option<String>, String> {
+	if tries == 0 {
+		return Ok(None);
+	}
+
+	// for each URL pair, try the HEAD request
+	let res = {
+		// for url base and host in URL_PAIRS, try reddit_short_head(path.clone(), true, url_base, url_base_host) and if it succeeds, set res. else, res = None
+		let mut res = None;
+		for (url_base, url_base_host) in URL_PAIRS {
+			res = reddit_short_head(path.clone(), true, url_base, url_base_host).await.ok();
+			if let Some(res) = &res {
+				if !res.status().is_client_error() {
+					break;
+				}
+			}
+		}
+		res
+	};
+
+	let res = res.ok_or_else(|| "Unable to make HEAD request to Reddit.".to_string())?;
 	let status = res.status().as_u16();
+	let policy_error = res.headers().get(header::RETRY_AFTER).is_some();
 
 	match status {
-		429 => Err("Too many requests.".to_string()),
-
 		// If Reddit responds with a 2xx, then the path is already canonical.
 		200..=299 => Ok(Some(path)),
 
@@ -73,6 +98,7 @@ pub async fn canonical_path(path: String) -> Result<Option<String>, String> {
 				let Ok(original) = val.to_str() else {
 					return Err("Unable to decode Location header.".to_string());
 				};
+
 				// We need to strip the .json suffix from the original path.
 				// In addition, we want to remove share parameters.
 				// Cut it off here instead of letting it propagate all the way
@@ -85,7 +111,9 @@ pub async fn canonical_path(path: String) -> Result<Option<String>, String> {
 				// also remove all Reddit domain parts with format_url.
 				// Otherwise, it will literally redirect to Reddit.com.
 				let uri = format_url(stripped_uri);
-				Ok(Some(uri))
+
+				// Decrement tries and try again
+				canonical_path(uri, tries - 1).await
 			}
 			None => Ok(None),
 		},
@@ -94,6 +122,12 @@ pub async fn canonical_path(path: String) -> Result<Option<String>, String> {
 		// as above), return a None.
 		300..=399 => Ok(None),
 
+		// Rate limiting
+		429 => Err("Too many requests.".to_string()),
+
+		// Special condition rate limiting - https://github.com/redlib-org/redlib/issues/229
+		403 if policy_error => Err("Too many requests.".to_string()),
+
 		_ => Ok(
 			res
 				.headers()
@@ -120,7 +154,7 @@ async fn stream(url: &str, req: &Request<Body>) -> Result<Response<Body>, String
 	let parsed_uri = url.parse::<Uri>().map_err(|_| "Couldn't parse URL".to_string())?;
 
 	// Build the hyper client from the HTTPS connector.
-	let client: Client<_, Body> = CLIENT.clone();
+	let client: &Lazy<Client<_, Body>> = &CLIENT;
 
 	let mut builder = Request::get(parsed_uri);
 
@@ -160,58 +194,62 @@ async fn stream(url: &str, req: &Request<Body>) -> Result<Response<Body>, String
 /// Makes a GET request to Reddit at `path`. By default, this will honor HTTP
 /// 3xx codes Reddit returns and will automatically redirect.
 fn reddit_get(path: String, quarantine: bool) -> Boxed<Result<Response<Body>, String>> {
-	request(&Method::GET, path, true, quarantine)
+	request(&Method::GET, path, true, quarantine, REDDIT_URL_BASE, REDDIT_URL_BASE_HOST)
 }
 
-/// Makes a HEAD request to Reddit at `path`. This will not follow redirects.
-fn reddit_head(path: String, quarantine: bool) -> Boxed<Result<Response<Body>, String>> {
-	request(&Method::HEAD, path, false, quarantine)
+/// Makes a HEAD request to Reddit at `path, using the short URL base. This will not follow redirects.
+fn reddit_short_head(path: String, quarantine: bool, base_path: &'static str, host: &'static str) -> Boxed<Result<Response<Body>, String>> {
+	request(&Method::HEAD, path, false, quarantine, base_path, host)
 }
 
+// /// Makes a HEAD request to Reddit at `path`. This will not follow redirects.
+// fn reddit_head(path: String, quarantine: bool) -> Boxed<Result<Response<Body>, String>> {
+// 	request(&Method::HEAD, path, false, quarantine, false)
+// }
+// Unused - reddit_head is only ever called in the context of a short URL
+
 /// Makes a request to Reddit. If `redirect` is `true`, `request_with_redirect`
 /// will recurse on the URL that Reddit provides in the Location HTTP header
 /// in its response.
-fn request(method: &'static Method, path: String, redirect: bool, quarantine: bool) -> Boxed<Result<Response<Body>, String>> {
+fn request(method: &'static Method, path: String, redirect: bool, quarantine: bool, base_path: &'static str, host: &'static str) -> Boxed<Result<Response<Body>, String>> {
 	// Build Reddit URL from path.
-	let url = format!("{REDDIT_URL_BASE}{path}");
+	let url = format!("{base_path}{path}");
 
 	// Construct the hyper client from the HTTPS connector.
-	let client: Client<_, Body> = CLIENT.clone();
-
-	let (token, vendor_id, device_id, user_agent, loid) = {
-		let client = OAUTH_CLIENT.load_full();
-		(
-			client.token.clone(),
-			client.headers_map.get("Client-Vendor-Id").cloned().unwrap_or_default(),
-			client.headers_map.get("X-Reddit-Device-Id").cloned().unwrap_or_default(),
-			client.headers_map.get("User-Agent").cloned().unwrap_or_default(),
-			client.headers_map.get("x-reddit-loid").cloned().unwrap_or_default(),
-		)
-	};
+	let client: &Lazy<Client<_, Body>> = &CLIENT;
 
 	// Build request to Reddit. When making a GET, request gzip compression.
 	// (Reddit doesn't do brotli yet.)
-	let builder = Request::builder()
-		.method(method)
-		.uri(&url)
-		.header("User-Agent", user_agent)
-		.header("Client-Vendor-Id", vendor_id)
-		.header("X-Reddit-Device-Id", device_id)
-		.header("x-reddit-loid", loid)
-		.header("Host", "oauth.reddit.com")
-		.header("Authorization", &format!("Bearer {token}"))
-		.header("Accept-Encoding", if method == Method::GET { "gzip" } else { "identity" })
-		.header("Accept-Language", "en-US,en;q=0.5")
-		.header("Connection", "keep-alive")
-		.header(
-			"Cookie",
+	let mut headers: Vec<(String, String)> = vec![
+		("Host".into(), host.into()),
+		("Accept-Encoding".into(), if method == Method::GET { "gzip".into() } else { "identity".into() }),
+		(
+			"Cookie".into(),
 			if quarantine {
-				"_options=%7B%22pref_quarantine_optin%22%3A%20true%2C%20%22pref_gated_sr_optin%22%3A%20true%7D"
+				"_options=%7B%22pref_quarantine_optin%22%3A%20true%2C%20%22pref_gated_sr_optin%22%3A%20true%7D".into()
 			} else {
-				""
+				"".into()
 			},
-		)
-		.body(Body::empty());
+		),
+	];
+
+	{
+		let client = OAUTH_CLIENT.load_full();
+		for (key, value) in client.headers_map.clone() {
+			headers.push((key, value));
+		}
+	}
+
+	// shuffle headers: https://github.com/redlib-org/redlib/issues/324
+	fastrand::shuffle(&mut headers);
+
+	let mut builder = Request::builder().method(method).uri(&url);
+
+	for (key, value) in headers {
+		builder = builder.header(key, value);
+	}
+
+	let builder = builder.body(Body::empty());
 
 	async move {
 		match builder {
@@ -224,7 +262,7 @@ fn request(method: &'static Method, path: String, redirect: bool, quarantine: bo
 							return Ok(response);
 						};
 						let location_header = response.headers().get(header::LOCATION);
-						if location_header == Some(&HeaderValue::from_static("https://www.reddit.com/")) {
+						if location_header == Some(&HeaderValue::from_static(ALTERNATIVE_REDDIT_URL_BASE)) {
 							return Err("Reddit response was invalid".to_string());
 						}
 						return request(
@@ -253,6 +291,8 @@ fn request(method: &'static Method, path: String, redirect: bool, quarantine: bo
 								.to_string(),
 							true,
 							quarantine,
+							base_path,
+							host,
 						)
 						.await;
 					};
@@ -348,6 +388,12 @@ pub async fn json(path: String, quarantine: bool) -> Result<Value, String> {
 					"Ratelimit remaining: Header says {remaining}, we have {current_rate_limit}. Resets in {reset}. Rollover: {}. Ratelimit used: {used}",
 					if is_rolling_over { "yes" } else { "no" },
 				);
+
+				// If can parse remaining as a float, round to a u16 and save
+				if let Ok(val) = remaining.parse::<f32>() {
+					OAUTH_RATELIMIT_REMAINING.store(val.round() as u16, Ordering::SeqCst);
+				}
+
 				Some(reset)
 			} else {
 				None
@@ -374,6 +420,16 @@ pub async fn json(path: String, quarantine: bool) -> Result<Value, String> {
 					match serde_json::from_reader(body.reader()) {
 						Ok(value) => {
 							let json: Value = value;
+
+							// If user is suspended
+							if let Some(data) = json.get("data") {
+								if let Some(is_suspended) = data.get("is_suspended").and_then(Value::as_bool) {
+									if is_suspended {
+										return Err("suspended".into());
+									}
+								}
+							}
+
 							// If Reddit returned an error
 							if json["error"].is_i64() {
 								// OAuth token has expired; http status 401
@@ -382,6 +438,24 @@ pub async fn json(path: String, quarantine: bool) -> Result<Value, String> {
 									let () = force_refresh_token().await;
 									return Err("OAuth token has expired. Please refresh the page!".to_string());
 								}
+
+								// Handle quarantined
+								if json["reason"] == "quarantined" {
+									return Err("quarantined".into());
+								}
+								// Handle gated
+								if json["reason"] == "gated" {
+									return Err("gated".into());
+								}
+								// Handle private subs
+								if json["reason"] == "private" {
+									return Err("private".into());
+								}
+								// Handle banned subs
+								if json["reason"] == "banned" {
+									return Err("banned".into());
+								}
+
 								Err(format!("Reddit error {} \"{}\": {} | {path}", json["error"], json["reason"], json["message"]))
 							} else {
 								Ok(json)
@@ -404,8 +478,57 @@ pub async fn json(path: String, quarantine: bool) -> Result<Value, String> {
 	}
 }
 
+async fn self_check(sub: &str) -> Result<(), String> {
+	let query = format!("/r/{sub}/hot.json?&raw_json=1");
+
+	match Post::fetch(&query, true).await {
+		Ok(_) => Ok(()),
+		Err(e) => Err(e),
+	}
+}
+
+pub async fn rate_limit_check() -> Result<(), String> {
+	// First, check a subreddit.
+	self_check("reddit").await?;
+	// This will reduce the rate limit to 99. Assert this check.
+	if OAUTH_RATELIMIT_REMAINING.load(Ordering::SeqCst) != 99 {
+		return Err(format!("Rate limit check failed: expected 99, got {}", OAUTH_RATELIMIT_REMAINING.load(Ordering::SeqCst)));
+	}
+	// Now, we switch out the OAuth client.
+	// This checks for the IP rate limit association.
+	force_refresh_token().await;
+	// Now, check a new sub to break cache.
+	self_check("rust").await?;
+	// Again, assert the rate limit check.
+	if OAUTH_RATELIMIT_REMAINING.load(Ordering::SeqCst) != 99 {
+		return Err(format!("Rate limit check failed: expected 99, got {}", OAUTH_RATELIMIT_REMAINING.load(Ordering::SeqCst)));
+	}
+
+	Ok(())
+}
+
 #[cfg(test)]
-static POPULAR_URL: &str = "/r/popular/hot.json?&raw_json=1&geo_filter=GLOBAL";
+use {crate::config::get_setting, sealed_test::prelude::*};
+
+#[tokio::test(flavor = "multi_thread")]
+async fn test_rate_limit_check() {
+	rate_limit_check().await.unwrap();
+}
+
+#[test]
+#[sealed_test(env = [("REDLIB_DEFAULT_SUBSCRIPTIONS", "rust")])]
+fn test_default_subscriptions() {
+	tokio::runtime::Builder::new_multi_thread().enable_all().build().unwrap().block_on(async {
+		let subscriptions = get_setting("REDLIB_DEFAULT_SUBSCRIPTIONS");
+		assert!(subscriptions.is_some());
+
+		// check rate limit
+		rate_limit_check().await.unwrap();
+	});
+}
+
+#[cfg(test)]
+const POPULAR_URL: &str = "/r/popular/hot.json?&raw_json=1&geo_filter=GLOBAL";
 
 #[tokio::test(flavor = "multi_thread")]
 async fn test_localization_popular() {
@@ -417,13 +540,28 @@ async fn test_localization_popular() {
 async fn test_obfuscated_share_link() {
 	let share_link = "/r/rust/s/kPgq8WNHRK".into();
 	// Correct link without share parameters
-	let canonical_link = "/r/rust/comments/18t5968/why_use_tuple_struct_over_standard_struct/kfbqlbc".into();
-	assert_eq!(canonical_path(share_link).await, Ok(Some(canonical_link)));
+	let canonical_link = "/r/rust/comments/18t5968/why_use_tuple_struct_over_standard_struct/kfbqlbc/".into();
+	assert_eq!(canonical_path(share_link, 3).await, Ok(Some(canonical_link)));
 }
 
 #[tokio::test(flavor = "multi_thread")]
-async fn test_share_link_strip_json() {
-	let link = "/17krzvz".into();
-	let canonical_link = "/r/nfl/comments/17krzvz/rapoport_sources_former_no_2_overall_pick/".into();
-	assert_eq!(canonical_path(link).await, Ok(Some(canonical_link)));
+async fn test_private_sub() {
+	let link = json("/r/suicide/about.json?raw_json=1".into(), true).await;
+	assert!(link.is_err());
+	assert_eq!(link, Err("private".into()));
+}
+
+#[tokio::test(flavor = "multi_thread")]
+async fn test_banned_sub() {
+	let link = json("/r/aaa/about.json?raw_json=1".into(), true).await;
+	assert!(link.is_err());
+	assert_eq!(link, Err("banned".into()));
+}
+
+#[tokio::test(flavor = "multi_thread")]
+async fn test_gated_sub() {
+	// quarantine to false to specifically catch when we _don't_ catch it
+	let link = json("/r/drugs/about.json?raw_json=1".into(), false).await;
+	assert!(link.is_err());
+	assert_eq!(link, Err("gated".into()));
 }
