@@ -1,20 +1,28 @@
 #![allow(dead_code)]
-use crate::config::get_setting;
+#![allow(clippy::cmp_owned)]
+
+use crate::config::{self, get_setting};
 //
 // CRATES
 //
 use crate::{client::json, server::RequestExt};
-use askama::Template;
 use cookie::Cookie;
 use hyper::{Body, Request, Response};
+use libflate::deflate::{Decoder, Encoder};
 use log::error;
 use once_cell::sync::Lazy;
 use regex::Regex;
+use revision::revisioned;
+use rinja::Template;
 use rust_embed::RustEmbed;
+use serde::{Deserialize, Deserializer, Serialize, Serializer};
 use serde_json::Value;
+use serde_json_path::{JsonPath, JsonPathExt};
 use std::collections::{HashMap, HashSet};
 use std::env;
+use std::io::{Read, Write};
 use std::str::FromStr;
+use std::string::ToString;
 use time::{macros::format_description, Duration, OffsetDateTime};
 use url::Url;
 
@@ -44,6 +52,7 @@ pub enum ResourceType {
 }
 
 // Post flair with content, background color and foreground color
+#[derive(Serialize)]
 pub struct Flair {
 	pub flair_parts: Vec<FlairPart>,
 	pub text: String,
@@ -52,7 +61,7 @@ pub struct Flair {
 }
 
 // Part of flair, either emoji or text
-#[derive(Clone)]
+#[derive(Clone, Serialize)]
 pub struct FlairPart {
 	pub flair_part_type: String,
 	pub value: String,
@@ -94,12 +103,14 @@ impl FlairPart {
 	}
 }
 
+#[derive(Serialize)]
 pub struct Author {
 	pub name: String,
 	pub flair: Flair,
 	pub distinguished: String,
 }
 
+#[derive(Serialize)]
 pub struct Poll {
 	pub poll_options: Vec<PollOption>,
 	pub voting_end_timestamp: (String, String),
@@ -127,6 +138,7 @@ impl Poll {
 	}
 }
 
+#[derive(Serialize)]
 pub struct PollOption {
 	pub id: u64,
 	pub text: String,
@@ -156,19 +168,21 @@ impl PollOption {
 }
 
 // Post flags with nsfw and stickied
+#[derive(Serialize)]
 pub struct Flags {
 	pub spoiler: bool,
 	pub nsfw: bool,
 	pub stickied: bool,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Serialize)]
 pub struct Media {
 	pub url: String,
 	pub alt_url: String,
 	pub width: i64,
 	pub height: i64,
 	pub poster: String,
+	pub download_name: String,
 }
 
 impl Media {
@@ -222,6 +236,14 @@ impl Media {
 			// If this post contains a gallery of images
 			gallery = GalleryMedia::parse(&data["gallery_data"]["items"], &data["media_metadata"]);
 
+			("gallery", &data["url"], None)
+		} else if data["crosspost_parent_list"][0]["is_gallery"].as_bool().unwrap_or_default() {
+			// If this post contains a gallery of images
+			gallery = GalleryMedia::parse(
+				&data["crosspost_parent_list"][0]["gallery_data"]["items"],
+				&data["crosspost_parent_list"][0]["media_metadata"],
+			);
+
 			("gallery", &data["url"], None)
 		} else if data["is_reddit_media_domain"].as_bool().unwrap_or_default() && data["domain"] == "i.redd.it" {
 			// If this post contains a reddit media (image) URL.
@@ -235,6 +257,15 @@ impl Media {
 
 		let alt_url = alt_url_val.map_or(String::new(), |val| format_url(val.as_str().unwrap_or_default()));
 
+		let download_name = if post_type == "image" || post_type == "gif" || post_type == "video" {
+			let permalink_base = url_path_basename(data["permalink"].as_str().unwrap_or_default());
+			let media_url_base = url_path_basename(url_val.as_str().unwrap_or_default());
+
+			format!("redlib_{permalink_base}_{media_url_base}")
+		} else {
+			String::new()
+		};
+
 		(
 			post_type.to_string(),
 			Self {
@@ -245,12 +276,14 @@ impl Media {
 				width: source["width"].as_i64().unwrap_or_default(),
 				height: source["height"].as_i64().unwrap_or_default(),
 				poster: format_url(source["url"].as_str().unwrap_or_default()),
+				download_name,
 			},
 			gallery,
 		)
 	}
 }
 
+#[derive(Serialize)]
 pub struct GalleryMedia {
 	pub url: String,
 	pub width: i64,
@@ -291,6 +324,7 @@ impl GalleryMedia {
 }
 
 // Post containing content, metadata and media
+#[derive(Serialize)]
 pub struct Post {
 	pub id: String,
 	pub title: String,
@@ -298,6 +332,7 @@ pub struct Post {
 	pub body: String,
 	pub author: Author,
 	pub permalink: String,
+	pub link_title: String,
 	pub poll: Option<Poll>,
 	pub score: (String, String),
 	pub upvote_ratio: i64,
@@ -315,6 +350,7 @@ pub struct Post {
 	pub gallery: Vec<GalleryMedia>,
 	pub awards: Awards,
 	pub nsfw: bool,
+	pub out_url: Option<String>,
 	pub ws_url: String,
 }
 
@@ -388,6 +424,7 @@ impl Post {
 					width: data["thumbnail_width"].as_i64().unwrap_or_default(),
 					height: data["thumbnail_height"].as_i64().unwrap_or_default(),
 					poster: String::new(),
+					download_name: String::new(),
 				},
 				media,
 				domain: val(post, "domain"),
@@ -411,6 +448,7 @@ impl Post {
 					stickied: data["stickied"].as_bool().unwrap_or_default() || data["pinned"].as_bool().unwrap_or_default(),
 				},
 				permalink: val(post, "permalink"),
+				link_title: val(post, "link_title"),
 				poll: Poll::parse(&data["poll_data"]),
 				rel_time,
 				created,
@@ -421,10 +459,9 @@ impl Post {
 				awards,
 				nsfw: post["data"]["over_18"].as_bool().unwrap_or_default(),
 				ws_url: val(post, "websocket_url"),
+				out_url: post["data"]["url_overridden_by_dest"].as_str().map(|a| a.to_string()),
 			});
 		}
-		posts.sort_by(|a, b| b.created_ts.cmp(&a.created_ts));
-		posts.sort_by(|a, b| b.flags.stickied.cmp(&a.flags.stickied));
 		Ok((posts, res["data"]["after"].as_str().unwrap_or_default().to_string()))
 	}
 }
@@ -454,7 +491,7 @@ pub struct Comment {
 	pub prefs: Preferences,
 }
 
-#[derive(Default, Clone)]
+#[derive(Default, Clone, Serialize)]
 pub struct Award {
 	pub name: String,
 	pub icon_url: String,
@@ -468,6 +505,7 @@ impl std::fmt::Display for Award {
 	}
 }
 
+#[derive(Serialize)]
 pub struct Awards(pub Vec<Award>);
 
 impl std::ops::Deref for Awards {
@@ -480,7 +518,7 @@ impl std::ops::Deref for Awards {
 
 impl std::fmt::Display for Awards {
 	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-		self.iter().fold(Ok(()), |result, award| result.and_then(|()| writeln!(f, "{award}")))
+		self.iter().try_fold((), |_, award| writeln!(f, "{award}"))
 	}
 }
 
@@ -515,6 +553,14 @@ pub struct ErrorTemplate {
 	pub url: String,
 }
 
+#[derive(Template)]
+#[template(path = "info.html")]
+pub struct InfoTemplate {
+	pub msg: String,
+	pub prefs: Preferences,
+	pub url: String,
+}
+
 /// Template for NSFW landing page. The landing page is displayed when a page's
 /// content is wholly NSFW, but a user has not enabled the option to view NSFW
 /// posts.
@@ -574,28 +620,74 @@ pub struct Params {
 	pub before: Option<String>,
 }
 
-#[derive(Default)]
+#[derive(Default, Serialize, Deserialize, Debug, PartialEq, Eq)]
+#[revisioned(revision = 1)]
 pub struct Preferences {
+	#[revision(start = 1)]
+	#[serde(skip_serializing, skip_deserializing)]
 	pub available_themes: Vec<String>,
+	#[revision(start = 1)]
 	pub theme: String,
+	#[revision(start = 1)]
 	pub front_page: String,
+	#[revision(start = 1)]
 	pub layout: String,
+	#[revision(start = 1)]
 	pub wide: String,
+	#[revision(start = 1)]
 	pub blur_spoiler: String,
+	#[revision(start = 1)]
 	pub show_nsfw: String,
+	#[revision(start = 1)]
 	pub blur_nsfw: String,
+	#[revision(start = 1)]
 	pub hide_hls_notification: String,
+	#[revision(start = 1)]
+	pub video_quality: String,
+	#[revision(start = 1)]
 	pub hide_sidebar_and_summary: String,
+	#[revision(start = 1)]
 	pub use_hls: String,
+	#[revision(start = 1)]
 	pub autoplay_videos: String,
+	#[revision(start = 1)]
 	pub fixed_navbar: String,
+	#[revision(start = 1)]
 	pub disable_visit_reddit_confirmation: String,
+	#[revision(start = 1)]
 	pub comment_sort: String,
+	#[revision(start = 1)]
 	pub post_sort: String,
+	#[revision(start = 1)]
+	#[serde(serialize_with = "serialize_vec_with_plus", deserialize_with = "deserialize_vec_with_plus")]
 	pub subscriptions: Vec<String>,
+	#[revision(start = 1)]
+	#[serde(serialize_with = "serialize_vec_with_plus", deserialize_with = "deserialize_vec_with_plus")]
 	pub filters: Vec<String>,
+	#[revision(start = 1)]
 	pub hide_awards: String,
+	#[revision(start = 1)]
 	pub hide_score: String,
+	#[revision(start = 1)]
+	pub remove_default_feeds: String,
+}
+
+fn serialize_vec_with_plus<S>(vec: &[String], serializer: S) -> Result<S::Ok, S::Error>
+where
+	S: Serializer,
+{
+	serializer.serialize_str(&vec.join("+"))
+}
+
+fn deserialize_vec_with_plus<'de, D>(deserializer: D) -> Result<Vec<String>, D::Error>
+where
+	D: Deserializer<'de>,
+{
+	let string = String::deserialize(deserializer)?;
+	if string.is_empty() {
+		return Ok(Vec::new());
+	}
+	Ok(string.split('+').map(|s| s.to_string()).collect())
 }
 
 #[derive(RustEmbed)]
@@ -625,6 +717,7 @@ impl Preferences {
 			blur_nsfw: setting(req, "blur_nsfw"),
 			use_hls: setting(req, "use_hls"),
 			hide_hls_notification: setting(req, "hide_hls_notification"),
+			video_quality: setting(req, "video_quality"),
 			autoplay_videos: setting(req, "autoplay_videos"),
 			fixed_navbar: setting_or_default(req, "fixed_navbar", "on".to_string()),
 			disable_visit_reddit_confirmation: setting(req, "disable_visit_reddit_confirmation"),
@@ -634,8 +727,36 @@ impl Preferences {
 			filters: setting(req, "filters").split('+').map(String::from).filter(|s| !s.is_empty()).collect(),
 			hide_awards: setting(req, "hide_awards"),
 			hide_score: setting(req, "hide_score"),
+			remove_default_feeds: setting(req, "remove_default_feeds"),
 		}
 	}
+
+	pub fn to_urlencoded(&self) -> Result<String, String> {
+		serde_urlencoded::to_string(self).map_err(|e| e.to_string())
+	}
+
+	pub fn to_bincode(&self) -> Result<Vec<u8>, String> {
+		bincode::serialize(self).map_err(|e| e.to_string())
+	}
+	pub fn to_compressed_bincode(&self) -> Result<Vec<u8>, String> {
+		deflate_compress(self.to_bincode()?)
+	}
+	pub fn to_bincode_str(&self) -> Result<String, String> {
+		Ok(base2048::encode(&self.to_compressed_bincode()?))
+	}
+}
+
+pub fn deflate_compress(i: Vec<u8>) -> Result<Vec<u8>, String> {
+	let mut e = Encoder::new(Vec::new());
+	e.write_all(&i).map_err(|e| e.to_string())?;
+	e.finish().into_result().map_err(|e| e.to_string())
+}
+
+pub fn deflate_decompress(i: Vec<u8>) -> Result<Vec<u8>, String> {
+	let mut decoder = Decoder::new(&i[..]);
+	let mut out = Vec::new();
+	decoder.read_to_end(&mut out).map_err(|e| format!("Failed to read from gzip decoder: {}", e))?;
+	Ok(out)
 }
 
 /// Gets a `HashSet` of filters from the cookie in the given `Request`.
@@ -691,7 +812,15 @@ pub async fn parse_post(post: &Value) -> Post {
 			get_setting("REDLIB_PUSHSHIFT_FRONTEND").unwrap_or_else(|| String::from(crate::config::DEFAULT_PUSHSHIFT_FRONTEND)),
 		)
 	} else {
-		rewrite_urls(&val(post, "selftext_html"))
+		let selftext = val(post, "selftext");
+		if selftext.contains("```") {
+			let mut html_output = String::new();
+			let parser = pulldown_cmark::Parser::new(&selftext);
+			pulldown_cmark::html::push_html(&mut html_output, parser);
+			rewrite_urls(&html_output)
+		} else {
+			rewrite_urls(&val(post, "selftext_html"))
+		}
 	};
 
 	// Build a post using data parsed from Reddit post API
@@ -715,6 +844,7 @@ pub async fn parse_post(post: &Value) -> Post {
 			distinguished: val(post, "distinguished"),
 		},
 		permalink,
+		link_title: val(post, "link_title"),
 		poll,
 		score: format_num(score),
 		upvote_ratio: ratio as i64,
@@ -726,6 +856,7 @@ pub async fn parse_post(post: &Value) -> Post {
 			width: post["data"]["thumbnail_width"].as_i64().unwrap_or_default(),
 			height: post["data"]["thumbnail_height"].as_i64().unwrap_or_default(),
 			poster: String::new(),
+			download_name: String::new(),
 		},
 		flair: Flair {
 			flair_parts: FlairPart::parse(
@@ -756,6 +887,7 @@ pub async fn parse_post(post: &Value) -> Post {
 		awards,
 		nsfw: post["data"]["over_18"].as_bool().unwrap_or_default(),
 		ws_url: val(post, "websocket_url"),
+		out_url: post["data"]["url_overridden_by_dest"].as_str().map(|a| a.to_string()),
 	}
 }
 
@@ -779,18 +911,72 @@ pub fn param(path: &str, value: &str) -> Option<String> {
 // Retrieve the value of a setting by name
 pub fn setting(req: &Request<Body>, name: &str) -> String {
 	// Parse a cookie value from request
-	req
-		.cookie(name)
-		.unwrap_or_else(|| {
-			// If there is no cookie for this setting, try receiving a default from the config
-			if let Some(default) = get_setting(&format!("REDLIB_DEFAULT_{}", name.to_uppercase())) {
-				Cookie::new(name, default)
-			} else {
-				Cookie::from(name)
-			}
-		})
-		.value()
-		.to_string()
+
+	// If this was called with "subscriptions" and the "subscriptions" cookie has a value
+	if name == "subscriptions" && req.cookie("subscriptions").is_some() {
+		// Create subscriptions string
+		let mut subscriptions = String::new();
+
+		// Default subscriptions cookie
+		if req.cookie("subscriptions").is_some() {
+			subscriptions.push_str(req.cookie("subscriptions").unwrap().value());
+		}
+
+		// Start with first numbered subscription cookie
+		let mut subscriptions_number = 1;
+
+		// While whatever subscriptionsNUMBER cookie we're looking at has a value
+		while req.cookie(&format!("subscriptions{}", subscriptions_number)).is_some() {
+			// Push whatever subscriptionsNUMBER cookie we're looking at into the subscriptions string
+			subscriptions.push_str(req.cookie(&format!("subscriptions{}", subscriptions_number)).unwrap().value());
+
+			// Increment subscription cookie number
+			subscriptions_number += 1;
+		}
+
+		// Return the subscriptions cookies as one large string
+		subscriptions
+	}
+	// If this was called with "filters" and the "filters" cookie has a value
+	else if name == "filters" && req.cookie("filters").is_some() {
+		// Create filters string
+		let mut filters = String::new();
+
+		// Default filters cookie
+		if req.cookie("filters").is_some() {
+			filters.push_str(req.cookie("filters").unwrap().value());
+		}
+
+		// Start with first numbered filters cookie
+		let mut filters_number = 1;
+
+		// While whatever filtersNUMBER cookie we're looking at has a value
+		while req.cookie(&format!("filters{}", filters_number)).is_some() {
+			// Push whatever filtersNUMBER cookie we're looking at into the filters string
+			filters.push_str(req.cookie(&format!("filters{}", filters_number)).unwrap().value());
+
+			// Increment filters cookie number
+			filters_number += 1;
+		}
+
+		// Return the filters cookies as one large string
+		filters
+	}
+	// The above two still come to this if there was no existing value
+	else {
+		req
+			.cookie(name)
+			.unwrap_or_else(|| {
+				// If there is no cookie for this setting, try receiving a default from the config
+				if let Some(default) = get_setting(&format!("REDLIB_DEFAULT_{}", name.to_uppercase())) {
+					Cookie::new(name, default)
+				} else {
+					Cookie::from(name)
+				}
+			})
+			.value()
+			.to_string()
+	}
 }
 
 // Retrieve the value of a setting by name or the default value
@@ -806,11 +992,12 @@ pub fn setting_or_default(req: &Request<Body>, name: &str, default: String) -> S
 // Detect and redirect in the event of a random subreddit
 pub async fn catch_random(sub: &str, additional: &str) -> Result<Response<Body>, String> {
 	if sub == "random" || sub == "randnsfw" {
-		let new_sub = json(format!("/r/{sub}/about.json?raw_json=1"), false).await?["data"]["display_name"]
-			.as_str()
-			.unwrap_or_default()
-			.to_string();
-		Ok(redirect(&format!("/r/{new_sub}{additional}")))
+		Ok(redirect(&format!(
+			"/r/{}{additional}",
+			json(format!("/r/{sub}/about.json?raw_json=1"), false).await?["data"]["display_name"]
+				.as_str()
+				.unwrap_or_default()
+		)))
 	} else {
 		Err("No redirect needed".to_string())
 	}
@@ -888,9 +1075,20 @@ pub fn format_url(url: &str) -> String {
 	}
 }
 
+static REGEX_BULLET: Lazy<Regex> = Lazy::new(|| Regex::new(r"(?m)^- (.*)$").unwrap());
+static REGEX_BULLET_CONSECUTIVE_LINES: Lazy<Regex> = Lazy::new(|| Regex::new(r"</ul>\n<ul>").unwrap());
+
+pub fn render_bullet_lists(input_text: &str) -> String {
+	// ref: https://stackoverflow.com/a/4902622
+	// First enclose each bullet with <ul> <li> tags
+	let text1 = REGEX_BULLET.replace_all(input_text, "<ul><li>$1</li></ul>").to_string();
+	// Then remove any consecutive </ul> <ul> tags
+	REGEX_BULLET_CONSECUTIVE_LINES.replace_all(&text1, "").to_string()
+}
+
 // These are links we want to replace in-body
 static REDDIT_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r#"href="(https|http|)://(www\.|old\.|np\.|amp\.|new\.|)(reddit\.com|redd\.it)/"#).unwrap());
-static REDDIT_PREVIEW_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r"https?://(external-preview|preview|i)\.redd\.it(.*)[^?]").unwrap());
+static REDDIT_PREVIEW_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r"https?://(external-preview|preview|i)\.redd\.it(.*)").unwrap());
 static REDDIT_EMOJI_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r"https?://(www|).redditstatic\.com/(.*)").unwrap());
 static REDLIB_PREVIEW_LINK_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r#"/(img|preview/)(pre|external-pre)?/(.*?)>"#).unwrap());
 static REDLIB_PREVIEW_TEXT_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r">(.*?)</a>").unwrap());
@@ -899,14 +1097,20 @@ static REDLIB_PREVIEW_TEXT_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r">(.*?)
 pub fn rewrite_urls(input_text: &str) -> String {
 	let mut text1 =
 		// Rewrite Reddit links to Redlib
-		REDDIT_REGEX.replace_all(input_text, r#"href="/"#)
-			.to_string();
-	text1 = REDDIT_EMOJI_REGEX
-		.replace_all(&text1, format_url(REDDIT_EMOJI_REGEX.find(&text1).map(|x| x.as_str()).unwrap_or_default()))
-		.to_string()
-		// Remove (html-encoded) "\" from URLs.
-		.replace("%5C", "")
-		.replace("\\_", "_");
+		REDDIT_REGEX.replace_all(input_text, r#"href="/"#).to_string();
+
+	loop {
+		if REDDIT_EMOJI_REGEX.find(&text1).is_none() {
+			break;
+		} else {
+			text1 = REDDIT_EMOJI_REGEX
+				.replace_all(&text1, format_url(REDDIT_EMOJI_REGEX.find(&text1).map(|x| x.as_str()).unwrap_or_default()))
+				.to_string()
+		}
+	}
+
+	// Remove (html-encoded) "\" from URLs.
+	text1 = text1.replace("%5C", "").replace("\\_", "_");
 
 	// Rewrite external media previews to Redlib
 	loop {
@@ -915,53 +1119,129 @@ pub fn rewrite_urls(input_text: &str) -> String {
 		} else {
 			let formatted_url = format_url(REDDIT_PREVIEW_REGEX.find(&text1).map(|x| x.as_str()).unwrap_or_default());
 
-			let image_url = REDLIB_PREVIEW_LINK_REGEX.find(&formatted_url).map_or("", |m| m.as_str()).to_string();
-			let mut image_caption = REDLIB_PREVIEW_TEXT_REGEX.find(&formatted_url).map_or("", |m| m.as_str()).to_string();
+			let image_url = REDLIB_PREVIEW_LINK_REGEX.find(&formatted_url).map_or("", |m| m.as_str());
+			let mut image_caption = REDLIB_PREVIEW_TEXT_REGEX.find(&formatted_url).map_or("", |m| m.as_str());
 
 			/* As long as image_caption isn't empty remove first and last four characters of image_text to leave us with just the text in the caption without any HTML.
 			This makes it possible to enclose it in a <figcaption> later on without having stray HTML breaking it */
 			if !image_caption.is_empty() {
-				image_caption = image_caption[1..image_caption.len() - 4].to_string();
+				image_caption = &image_caption[1..image_caption.len() - 4];
 			}
 
 			// image_url contains > at the end of it, and right above this we remove image_text's front >, leaving us with just a single > between them
-			let image_to_replace = format!("<a href=\"{image_url}{image_caption}</a>");
-
-			// _image_replacement needs to be in scope for the replacement at the bottom of the loop
-			let mut _image_replacement = String::new();
+			let image_to_replace = format!("<p><a href=\"{image_url}{image_caption}</a></p>");
 
 			/* We don't want to show a caption that's just the image's link, so we check if we find a Reddit preview link within the image's caption.
 			If we don't find one we must have actual text, so we include a <figcaption> block that contains it.
 			Otherwise we don't include the <figcaption> block as we don't need it. */
-			if REDDIT_PREVIEW_REGEX.find(&image_caption).is_none() {
+			let _image_replacement = if REDDIT_PREVIEW_REGEX.find(image_caption).is_none() {
 				// Without this " would show as \" instead. "\&quot;" is how the quotes are formatted within image_text beforehand
-				image_caption = image_caption.replace("\\&quot;", "\"");
-
-				_image_replacement = format!("<figure><a href=\"{image_url}<img loading=\"lazy\" src=\"{image_url}</a><figcaption>{image_caption}</figcaption></figure>");
+				format!(
+					"<figure><a href=\"{image_url}<img loading=\"lazy\" src=\"{image_url}</a><figcaption>{}</figcaption></figure>",
+					image_caption.replace("\\&quot;", "\"")
+				)
 			} else {
-				_image_replacement = format!("<figure><a href=\"{image_url}<img loading=\"lazy\" src=\"{image_url}</a></figure>");
-			}
+				format!("<figure><a href=\"{image_url}<img loading=\"lazy\" src=\"{image_url}</a></figure>")
+			};
 
 			/* In order to know if we're dealing with a normal or external preview we need to take a look at the first capture group of REDDIT_PREVIEW_REGEX
 			if it's preview we're dealing with something that needs /preview/pre, external-preview is /preview/external-pre, and i is /img */
-			let reddit_preview_regex_capture = REDDIT_PREVIEW_REGEX.captures(&text1).unwrap().get(1).map_or("", |m| m.as_str()).to_string();
-			let mut _preview_type = String::new();
-			if reddit_preview_regex_capture == "preview" {
-				_preview_type = "/preview/pre".to_string();
-			} else if reddit_preview_regex_capture == "external-preview" {
-				_preview_type = "/preview/external-pre".to_string();
-			} else {
-				_preview_type = "/img".to_string();
-			}
+			let reddit_preview_regex_capture = REDDIT_PREVIEW_REGEX.captures(&text1).unwrap().get(1).map_or("", |m| m.as_str());
+
+			let _preview_type = match reddit_preview_regex_capture {
+				"preview" => "/preview/pre",
+				"external-preview" => "/preview/external-pre",
+				_ => "/img",
+			};
 
 			text1 = REDDIT_PREVIEW_REGEX
 				.replace(&text1, format!("{_preview_type}$2"))
 				.replace(&image_to_replace, &_image_replacement)
-				.to_string()
 		}
 	}
 }
 
+// These links all follow a pattern of "https://reddit-econ-prod-assets-permanent.s3.amazonaws.com/asset-manager/SUBREDDIT_ID/RANDOM_FILENAME.png"
+static REDDIT_EMOTE_LINK_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r#"https://reddit-econ-prod-assets-permanent.s3.amazonaws.com/asset-manager/(.*)"#).unwrap());
+
+// These all follow a pattern of '"emote|SUBREDDIT_IT|NUMBER"', we want the number
+static REDDIT_EMOTE_ID_NUMBER_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r#""emote\|.*\|(.*)""#).unwrap());
+
+pub fn rewrite_emotes(media_metadata: &Value, comment: String) -> String {
+	/* Create the paths we'll use to look for our data inside the json.
+	Because we don't know the name of any given emote we use a wildcard to parse them. */
+	let link_path = JsonPath::parse("$[*].s.u").expect("valid JSON Path");
+	let id_path = JsonPath::parse("$[*].id").expect("valid JSON Path");
+	let size_path = JsonPath::parse("$[*].s.y").expect("valid JSON Path");
+
+	// Extract all of the results from those json paths
+	let link_nodes = media_metadata.json_path(&link_path);
+	let id_nodes = media_metadata.json_path(&id_path);
+
+	// Initialize our vectors
+	let mut id_vec = Vec::new();
+	let mut link_vec = Vec::new();
+
+	// Add the relevant data to each of our vectors so we can access it by number later
+	for current_id in id_nodes {
+		id_vec.push(current_id)
+	}
+	for current_link in link_nodes {
+		link_vec.push(current_link)
+	}
+
+	/* Set index to the length of link_vec.
+	This is one larger than we'll actually be looking at, but we correct that later */
+	let mut index = link_vec.len();
+
+	// Comment needs to be in scope for when we call rewrite_urls()
+	let mut comment = comment;
+
+	/* Loop until index hits zero.
+	This also prevents us from trying to do anything on an empty vector */
+	while index != 0 {
+		/* Subtract 1 from index to get the real index we should be looking at.
+		Then continue on each subsequent loop to continue until we hit the last entry in the vector.
+		This is how we get this to deal with multiple emotes in a single message and properly replace each ID with it's link */
+		index -= 1;
+
+		// Convert our current index in id_vec into a string so we can search through it with regex
+		let current_id = id_vec[index].to_string();
+
+		/* The ID number can be multiple lengths, so we capture it with regex.
+		We also want to only attempt anything when we get matches to avoid panicking */
+		if let Some(id_capture) = REDDIT_EMOTE_ID_NUMBER_REGEX.captures(&current_id) {
+			// Format the ID to include the colons it has in the comment text
+			let id = format!(":{}:", &id_capture[1]);
+
+			// Convert current link to string to search through it with the regex
+			let link = link_vec[index].to_string();
+
+			// Make sure we only do operations when we get matches, otherwise we panic when trying to access the first match
+			if let Some(link_capture) = REDDIT_EMOTE_LINK_REGEX.captures(&link) {
+				/* Reddit sends a size for the image based on whether it's alone or accompanied by text.
+				It's a good idea and makes everything look nicer, so we'll do the same. */
+				let size = media_metadata.json_path(&size_path).first().unwrap().to_string();
+
+				// Replace the ID we found earlier in the comment with the respective image and it's link from the regex capture
+				let to_replace_with = format!(
+					"<img loading=\"lazy\" src=\"/emote/{} width=\"{size}\" height=\"{size}\" style=\"vertical-align:text-bottom\">",
+					&link_capture[1]
+				);
+
+				// Inside the comment replace the ID we found with the string that will embed the image
+				comment = comment.replace(&id, &to_replace_with);
+			}
+		}
+	}
+
+	// render bullet (unordered) lists
+	comment = render_bullet_lists(&comment);
+
+	// Call rewrite_urls() to transform any other Reddit links
+	rewrite_urls(&comment)
+}
+
 // Format vote count to a string that will be displayed.
 // Append `m` and `k` for millions and thousands respectively, and
 // round to the nearest tenth.
@@ -1054,6 +1334,20 @@ pub async fn error(req: Request<Body>, msg: &str) -> Result<Response<Body>, Stri
 	Ok(Response::builder().status(404).header("content-type", "text/html").body(body.into()).unwrap_or_default())
 }
 
+/// Renders a generic info landing page.
+pub async fn info(req: Request<Body>, msg: &str) -> Result<Response<Body>, String> {
+	let url = req.uri().to_string();
+	let body = InfoTemplate {
+		msg: msg.to_string(),
+		prefs: Preferences::new(&req),
+		url,
+	}
+	.render()
+	.unwrap_or_default();
+
+	Ok(Response::builder().status(200).header("content-type", "text/html").body(body.into()).unwrap_or_default())
+}
+
 /// Returns true if the config/env variable `REDLIB_SFW_ONLY` carries the
 /// value `on`.
 ///
@@ -1068,6 +1362,28 @@ pub fn sfw_only() -> bool {
 	}
 }
 
+/// Returns true if the config/env variable REDLIB_ENABLE_RSS is set to "on".
+/// If this variable is set as such, the instance will enable RSS feeds.
+/// Otherwise, the instance will not provide RSS feeds.
+pub fn enable_rss() -> bool {
+	match get_setting("REDLIB_ENABLE_RSS") {
+		Some(val) => val == "on",
+		None => false,
+	}
+}
+
+/// Returns true if the config/env variable `REDLIB_ROBOTS_DISABLE_INDEXING` carries the
+/// value `on`.
+///
+/// If this variable is set as such, the instance will block all robots in robots.txt and
+/// insert the noindex, nofollow meta tag on every page.
+pub fn disable_indexing() -> bool {
+	match get_setting("REDLIB_ROBOTS_DISABLE_INDEXING") {
+		Some(val) => val == "on",
+		None => false,
+	}
+}
+
 // Determines if a request shoud redirect to a nsfw landing gate.
 pub fn should_be_nsfw_gated(req: &Request<Body>, req_url: &str) -> bool {
 	let sfw_instance = sfw_only();
@@ -1109,9 +1425,37 @@ pub async fn nsfw_landing(req: Request<Body>, req_url: String) -> Result<Respons
 	Ok(Response::builder().status(403).header("content-type", "text/html").body(body.into()).unwrap_or_default())
 }
 
+// Returns the last (non-empty) segment of a path string
+pub fn url_path_basename(path: &str) -> String {
+	let url_result = Url::parse(format!("https://libredd.it/{path}").as_str());
+
+	if url_result.is_err() {
+		path.to_string()
+	} else {
+		let mut url = url_result.unwrap();
+		url.path_segments_mut().unwrap().pop_if_empty();
+
+		url.path_segments().unwrap().next_back().unwrap().to_string()
+	}
+}
+
+// Returns the URL of a post, as needed by RSS feeds
+pub fn get_post_url(post: &Post) -> String {
+	if let Some(out_url) = &post.out_url {
+		// Handle cross post
+		if out_url.starts_with("/r/") {
+			format!("{}{}", config::get_setting("REDLIB_FULL_URL").unwrap_or_default(), out_url)
+		} else {
+			out_url.to_string()
+		}
+	} else {
+		format!("{}{}", config::get_setting("REDLIB_FULL_URL").unwrap_or_default(), post.permalink)
+	}
+}
+
 #[cfg(test)]
 mod tests {
-	use super::{format_num, format_url, rewrite_urls};
+	use super::{format_num, format_url, rewrite_urls, Preferences};
 
 	#[test]
 	fn format_num_works() {
@@ -1178,6 +1522,36 @@ mod tests {
 		assert_eq!(format_url("nsfw"), "");
 		assert_eq!(format_url("spoiler"), "");
 	}
+	#[test]
+	fn serialize_prefs() {
+		let prefs = Preferences {
+			available_themes: vec![],
+			theme: "laserwave".to_owned(),
+			front_page: "default".to_owned(),
+			layout: "compact".to_owned(),
+			wide: "on".to_owned(),
+			blur_spoiler: "on".to_owned(),
+			show_nsfw: "off".to_owned(),
+			blur_nsfw: "on".to_owned(),
+			hide_hls_notification: "off".to_owned(),
+			video_quality: "best".to_owned(),
+			hide_sidebar_and_summary: "off".to_owned(),
+			use_hls: "on".to_owned(),
+			autoplay_videos: "on".to_owned(),
+			fixed_navbar: "on".to_owned(),
+			disable_visit_reddit_confirmation: "on".to_owned(),
+			comment_sort: "confidence".to_owned(),
+			post_sort: "top".to_owned(),
+			subscriptions: vec!["memes".to_owned(), "mildlyinteresting".to_owned()],
+			filters: vec![],
+			hide_awards: "off".to_owned(),
+			hide_score: "off".to_owned(),
+			remove_default_feeds: "off".to_owned(),
+		};
+		let urlencoded = serde_urlencoded::to_string(prefs).expect("Failed to serialize Prefs");
+
+		assert_eq!(urlencoded, "theme=laserwave&front_page=default&layout=compact&wide=on&blur_spoiler=on&show_nsfw=off&blur_nsfw=on&hide_hls_notification=off&video_quality=best&hide_sidebar_and_summary=off&use_hls=on&autoplay_videos=on&fixed_navbar=on&disable_visit_reddit_confirmation=on&comment_sort=confidence&post_sort=top&subscriptions=memes%2Bmildlyinteresting&filters=&hide_awards=off&hide_score=off&remove_default_feeds=off");
+	}
 }
 
 #[test]
@@ -1196,7 +1570,10 @@ async fn test_fetching_subreddit_quarantined() {
 
 #[tokio::test(flavor = "multi_thread")]
 async fn test_fetching_nsfw_subreddit() {
-	let subreddit = Post::fetch("/r/randnsfw", false).await;
+	// Gonwild is a place for closed, Euclidean Geometric shapes to exchange their nth terms for karma; showing off their edges in a comfortable environment without pressure.
+	// Find a good sub that is tagged NSFW but that actually isn't in case my future employers are watching (they probably are)
+	// switched from randnsfw as it is no longer functional.
+	let subreddit = Post::fetch("/r/gonwild", false).await;
 	assert!(subreddit.is_ok());
 	assert!(!subreddit.unwrap().0.is_empty());
 }
@@ -1214,6 +1591,104 @@ async fn test_fetching_ws() {
 fn test_rewriting_image_links() {
 	let input =
 		r#"<p><a href="https://preview.redd.it/6awags382xo31.png?width=2560&amp;format=png&amp;auto=webp&amp;s=9c563aed4f07a91bdd249b5a3cea43a79710dcfc">caption 1</a></p>"#;
-	let output = r#"<p><figure><a href="/preview/pre/6awags382xo31.png?width=2560&amp;format=png&amp;auto=webp&amp;s=9c563aed4f07a91bdd249b5a3cea43a79710dcfc"><img loading="lazy" src="/preview/pre/6awags382xo31.png?width=2560&amp;format=png&amp;auto=webp&amp;s=9c563aed4f07a91bdd249b5a3cea43a79710dcfc"></a><figcaption>caption 1</figcaption></figure></p"#;
+	let output = r#"<figure><a href="/preview/pre/6awags382xo31.png?width=2560&amp;format=png&amp;auto=webp&amp;s=9c563aed4f07a91bdd249b5a3cea43a79710dcfc"><img loading="lazy" src="/preview/pre/6awags382xo31.png?width=2560&amp;format=png&amp;auto=webp&amp;s=9c563aed4f07a91bdd249b5a3cea43a79710dcfc"></a><figcaption>caption 1</figcaption></figure>"#;
 	assert_eq!(rewrite_urls(input), output);
 }
+
+#[test]
+fn test_url_path_basename() {
+	// without trailing slash
+	assert_eq!(url_path_basename("/first/last"), "last");
+	// with trailing slash
+	assert_eq!(url_path_basename("/first/last/"), "last");
+	// with query parameters
+	assert_eq!(url_path_basename("/first/last/?some=query"), "last");
+	// file path
+	assert_eq!(url_path_basename("/cdn/image.jpg"), "image.jpg");
+	// when a full url is passed instead of just a path
+	assert_eq!(url_path_basename("https://doma.in/first/last"), "last");
+	// empty path
+	assert_eq!(url_path_basename("/"), "");
+}
+
+#[test]
+fn test_rewriting_emotes() {
+	let json_input = serde_json::from_str(r#"{"emote|t5_31hpy|2028":{"e":"Image","id":"emote|t5_31hpy|2028","m":"image/png","s":{"u":"https://reddit-econ-prod-assets-permanent.s3.amazonaws.com/asset-manager/t5_31hpy/PW6WsOaLcd.png","x":60,"y":60},"status":"valid","t":"sticker"}}"#).expect("Valid JSON");
+	let comment_input = r#"<div class="comment_body "><div class="md"><p>:2028:</p></div></div>"#;
+	let output = r#"<div class="comment_body "><div class="md"><p><img loading="lazy" src="/emote/t5_31hpy/PW6WsOaLcd.png" width="60" height="60" style="vertical-align:text-bottom"></p></div></div>"#;
+	assert_eq!(rewrite_emotes(&json_input, comment_input.to_string()), output);
+}
+
+#[test]
+fn test_rewriting_bullet_list() {
+	let input = r#"<div class="md"><p>Hi, I&#39;ve bought this very same monitor and found no calibration whatsoever. I have an ICC profile that has been set up since I&#39;ve installed its driver from the LG website and it works ok. I also used <a href="http://www.lagom.nl/lcd-test/">http://www.lagom.nl/lcd-test/</a> to calibrate it. After some good tinkering I&#39;ve found the following settings + the color profile from the driver gets me past all the tests perfectly:
+- Brightness 50 (still have to settle on this one, it&#39;s personal preference, it controls the backlight, not the colors)
+- Contrast 70 (which for me was the default one)
+- Picture mode Custom
+- Super resolution + Off (it looks horrible anyway)
+- Sharpness 50 (default one I think)
+- Black level High (low messes up gray colors)
+- DFC Off 
+- Response Time Middle (personal preference, <a href="https://www.blurbusters.com/">https://www.blurbusters.com/</a> show horrible overdrive with it on high)
+- Freesync doesn&#39;t matter
+- Black stabilizer 50
+- Gamma setting on 0 
+- Color Temp Medium
+How`s your monitor by the way? Any IPS bleed whatsoever? I either got lucky or the panel is pretty good, 0 bleed for me, just the usual IPS glow. How about the pixels? I see the pixels even at one meter away, especially on Microsoft Edge&#39;s icon for example, the blue background is just blocky, don&#39;t know why.</p>
+</div>"#;
+	let output = r#"<div class="md"><p>Hi, I&#39;ve bought this very same monitor and found no calibration whatsoever. I have an ICC profile that has been set up since I&#39;ve installed its driver from the LG website and it works ok. I also used <a href="http://www.lagom.nl/lcd-test/">http://www.lagom.nl/lcd-test/</a> to calibrate it. After some good tinkering I&#39;ve found the following settings + the color profile from the driver gets me past all the tests perfectly:
+<ul><li>Brightness 50 (still have to settle on this one, it&#39;s personal preference, it controls the backlight, not the colors)</li><li>Contrast 70 (which for me was the default one)</li><li>Picture mode Custom</li><li>Super resolution + Off (it looks horrible anyway)</li><li>Sharpness 50 (default one I think)</li><li>Black level High (low messes up gray colors)</li><li>DFC Off </li><li>Response Time Middle (personal preference, <a href="https://www.blurbusters.com/">https://www.blurbusters.com/</a> show horrible overdrive with it on high)</li><li>Freesync doesn&#39;t matter</li><li>Black stabilizer 50</li><li>Gamma setting on 0 </li><li>Color Temp Medium</li></ul>
+How`s your monitor by the way? Any IPS bleed whatsoever? I either got lucky or the panel is pretty good, 0 bleed for me, just the usual IPS glow. How about the pixels? I see the pixels even at one meter away, especially on Microsoft Edge&#39;s icon for example, the blue background is just blocky, don&#39;t know why.</p>
+</div>"#;
+
+	assert_eq!(render_bullet_lists(input), output);
+}
+
+#[test]
+fn test_default_prefs_serialization_loop_json() {
+	let prefs = Preferences::default();
+	let serialized = serde_json::to_string(&prefs).unwrap();
+	let deserialized: Preferences = serde_json::from_str(&serialized).unwrap();
+	assert_eq!(prefs, deserialized);
+}
+
+#[test]
+fn test_default_prefs_serialization_loop_bincode() {
+	let prefs = Preferences::default();
+	test_round_trip(&prefs, false);
+	test_round_trip(&prefs, true);
+}
+
+static KNOWN_GOOD_CONFIGS: &[&str] = &[
+	"ఴӅβØØҞÉဏႢձĬ༧ȒʯऌԔӵ୮༏",
+	"ਧՊΥÀÃǎƱГ۸ඣമĖฤ႙ʟาúໜϾௐɥঀĜໃહཞઠѫҲɂఙ࿔ǲઉƲӟӻĻฅΜδ໖ԜǗဖငƦơ৶Ą௩ԹʛใЛʃශаΏ",
+	"ਧԩΥÀÃÎŠ౭൩ඔႠϼҭöҪƸռઇԾॐნɔາǒՍҰच௨ಖມŃЉŐདƦ๙ϩএఠȝഽйʮჯඒϰळՋ௮ສ৵ऎΦѧਹಧଟƙŃ३î༦ŌပղयƟแҜ།",
+];
+
+#[test]
+fn test_known_good_configs_deserialization() {
+	for config in KNOWN_GOOD_CONFIGS {
+		let bytes = base2048::decode(config).unwrap();
+		let decompressed = deflate_decompress(bytes).unwrap();
+		assert!(bincode::deserialize::<Preferences>(&decompressed).is_ok());
+	}
+}
+
+#[test]
+fn test_known_good_configs_full_round_trip() {
+	for config in KNOWN_GOOD_CONFIGS {
+		let bytes = base2048::decode(config).unwrap();
+		let decompressed = deflate_decompress(bytes).unwrap();
+		let prefs: Preferences = bincode::deserialize(&decompressed).unwrap();
+		test_round_trip(&prefs, false);
+		test_round_trip(&prefs, true);
+	}
+}
+
+fn test_round_trip(input: &Preferences, compression: bool) {
+	let serialized = bincode::serialize(input).unwrap();
+	let compressed = if compression { deflate_compress(serialized).unwrap() } else { serialized };
+	let decompressed = if compression { deflate_decompress(compressed).unwrap() } else { compressed };
+	let deserialized: Preferences = bincode::deserialize(&decompressed).unwrap();
+	assert_eq!(*input, deserialized);
+}
